import configparser
import os
from typing import Any, Iterator, Optional, Sequence, Type, TypeVar

from pydantic import BaseModel
from snowflake.snowpark.row import Row
from snowflake.snowpark import Session
from threading import local

threadlocal_data = local()

AnyModel = TypeVar("AnyModel", bound=BaseModel)


class Database:
    def __init__(self) -> None:
        self.config = {}
        config = configparser.ConfigParser()
        config.read(os.path.expanduser("~/.snowsql/config"))
        self.config = dict(
            account=os.getenv("SNOWFLAKE_ACCOUNT")
            or config["connections"]["accountname"],
            user=os.getenv("SNOWFLAKE_USER") or config["connections"]["username"],
            password=os.getenv("SNOWFLAKE_PASSWORD")
            or config["connections"]["password"],
            warehouse=os.getenv("SNOWFLAKE_WAREHOUSE")
            or config["connections"].get("warehousename"),
            database=os.getenv("SNOWFLAKE_DATABASE")
            or config["connections"].get("dbname"),
            schema=os.getenv("SNOWFLAKE_SCHEMA")
            or config["connections"].get("schemaname"),
            role=os.getenv("SNOWFLAKE_ROLE") or config["connections"].get("rolename"),
        )
        session_builder = Session.builder.config("connection_name", "default")
        session_builder.configs(self.config)
        self.session = session_builder.create()

    @classmethod
    def get_instance(self) -> "Database":
        if not hasattr(threadlocal_data, "database_instance"):
            threadlocal_data.database_instance = Database()
        return threadlocal_data.database_instance

    def get_rows(
        self, sql: str, params: Optional[Sequence[Any]] = None
    ) -> Iterator[Row]:
        return self.session.sql(sql, params).collect()

    def query(
        self,
        model: Type[AnyModel],
        sql: str,
        params: Optional[Sequence[Any]] = None,
    ) -> Iterator[AnyModel]:
        for row in self.session.sql(sql, params).collect():
            yield model.model_validate(row.as_dict())



import sys
from time import sleep
from typing import Annotated, List, Optional
import typer

from blackneedles.console import print_table
from blackneedles.models.service import Service

app = typer.Typer(short_help="Manage services in Snowpark Container Services.")


@app.command("list")
def list_all_services(
    ctx: typer.Context,
    schema: Annotated[
        Optional[str], typer.Option(help="The schema to list services from")
    ] = None,
    compute_pool: Annotated[
        Optional[str],
        typer.Option("--compute-pool", help="The compute pool to list services from"),
    ] = None,
):
    """Gives the list of services and its current status"""
    services = Service.objects.from_namespace(
        {
            "schema_name": schema,
            "compute_pool": compute_pool,
        }
    )
    services = list(services)

    print_table(
        ctx,
        services,
        [
            "full_name",
            "name",
            "schema_name",
            "owner",
            "compute_pool",
            lambda s: str(s.min_instances),
            lambda s: str(s.max_instances),
            "status",
        ],
        ["ID", "Name", "Schema", "Owner", "Compute pool", "Min", "Max", "Status"],
    )


@app.command("describe")
def describe_service(
    ctx: typer.Context,
    service_names: Annotated[
        List[str], typer.Argument(..., help="The name of the service to describe")
    ],
):
    """Describe a given service name."""
    services = [Service.objects.get(name) for name in service_names]
    print_table(
        ctx,
        services,
        [
            "full_path",
            "compute_pool",
            "status",
            "spec",
        ],
        [
            "ID",
            "Compute pool" "Status",
            "spec",
        ],
    )


@app.command("logs")
def get_service_logs(
    ctx: typer.Context,
    service_name: Annotated[
        str, typer.Argument(..., help="The name of the service to describe")
    ],
    instance_id: Annotated[
        str,
        typer.Argument(
            ..., help="The service instance id to get logs from. If unsure, try '0'"
        ),
    ],
    container_name: Annotated[
        str, typer.Argument(..., help="The name of the container in the service")
    ],
    follow: Annotated[bool, typer.Option("-f")] = False,
):
    """
    Shows logs from the given service.

    If -f is passed, it will follow the logs.
    """
    service = Service.objects.get(service_name)
    log = service.get_logs(instance_id, container_name)
    if not follow:
        print(log)
        sys.exit(0)
    while True:
        if log:
            print(log)
        sleep(0.5)
        log = service.get_logs(instance_id, container_name).replace(log, "")


@app.command("alter")
def alter_status(
    ctx: typer.Context,
    service_name: Annotated[
        str, typer.Argument(..., help="The name of the service to describe")
    ],
    action: Annotated[
        str, typer.Argument(..., help="Service state to change [SUSPEND, RESUME]")
    ],
):
    """Changes the status of a given service."""
    service = Service.objects.get(service_name)
    service.alter_status(action)
    typer.echo(f"Service {service_name} is now {action}")



from dataclasses import dataclass
import os
from typing import Annotated, Literal, Optional
import typer
import sqlparse  # type: ignore

from blackneedles.commands import compute_pool, service
from blackneedles.connection import Database


@dataclass
class GlobalParams:
    table_style: Literal["rich", "plain"]


app = typer.Typer()
app.add_typer(service.app, name="service")
app.add_typer(compute_pool.app, name="compute-pool")


@app.command("install")
def install_procedures(
    ctx: typer.Context,
    grant_target: Annotated[
        Optional[str], typer.Option(help="To whom should we grant permission?")
    ] = None,
):
    """Installs in database the procedures needed for monitoring services."""
    here = os.path.dirname(__file__)
    db = Database.get_instance()
    procedures_file = os.path.join(here, "procedures", "monitoring_procedures.sql")
    with open(procedures_file) as f:
        sql = f.read()
    for statement in sqlparse.split(sql):
        print(f"** Executing the following statement:\n{statement}")
        print(db.get_rows(statement))

    if grant_target is not None:
        print(f"Granting permissions: {grant_target}")
        grant_file = os.path.join(here, "procedures", "grant_permissions.sql")
        with open(grant_file) as f:
            sql = f.read()
        for statement in sqlparse.split(sql):
            statement = statement.format(grant_target=grant_target)
            print(f"** Executing the following statement:\n{statement}")
            print(db.get_rows(statement))


@app.command("uninstall")
def uninstall_procedures(
    ctx: typer.Context,
    grant_target: Annotated[
        Optional[str], typer.Option(help="To whom should we grant permission?")
    ] = None,
):
    """Removes from snowflake any procedures created by the install command."""
    here = os.path.dirname(__file__)
    db = Database.get_instance()
    procedures_file = os.path.join(here, "procedures", "uninstall.sql")
    with open(procedures_file) as f:
        sql = f.read()
    for statement in sqlparse.split(sql):
        print(f"** Executing the following statement:\n{statement}")
        print(db.get_rows(statement))


@app.callback()
def global_options(
    ctx: typer.Context,
    table_style: Annotated[
        str, typer.Option("--table-style", help="Output table style")
    ] = "rich",
):
    if ctx.resilient_parsing:
        return
    if table_style not in ("rich", "plain"):
        raise typer.BadParameter("table-style must be either 'rich' or 'plain'")
    ctx.obj = GlobalParams(
        table_style=table_style,  # type: ignore
    )
